package function.builtin.special;

import function.ArgumentValidator;
import function.FunctionNames;
import function.LispSpecialFunction;
import recursion.TailCall;
import sexpression.Cons;
import sexpression.Nil;
import sexpression.SExpression;
import sexpression.Symbol;

import static function.builtin.Eval.eval;
import static function.builtin.predicate.EQUAL.isEqual;
import static recursion.TailCalls.done;
import static recursion.TailCalls.tailCall;

@FunctionNames({ "CASE" })
public class CASE extends LispSpecialFunction {

    private ArgumentValidator argumentValidator;

    public CASE(String name) {
        this.argumentValidator = new ArgumentValidator(name);
        this.argumentValidator.setMinimumNumberOfArguments(1);
        this.argumentValidator.setTrailingArgumentExpectedType(Cons.class);
        this.argumentValidator.setTrailingArgumentExcludedType(Nil.class);
    }

    @Override
    public SExpression call(Cons argumentList) {
        argumentValidator.validate(argumentList);
        SExpression key = eval(argumentList.getFirst());

        return callTailRecursive(key, (Cons) argumentList.getRest()).invoke();
    }

    private TailCall<SExpression> callTailRecursive(SExpression key, Cons argumentList) {
        if (argumentList.isNull())
            return done(Nil.INSTANCE);

        Cons clause = (Cons) argumentList.getFirst();
        Cons remainingClauses = (Cons) argumentList.getRest();
        SExpression keyList = clause.getFirst();

        if (isMatch(key, keyList))
            return done(evaluateConsequents(clause.getRest()));

        return tailCall(() -> callTailRecursive(key, remainingClauses));
    }

    private boolean isMatch(SExpression key, SExpression keyList) {
        if (keyList.isNull())
            return false;
        else if (keyList.isCons())
            return containsMatch(key, keyList);

        return isEqual(key, keyList) || isEqual(Symbol.Companion.getT(), keyList);
    }

    private boolean containsMatch(SExpression key, SExpression keyList) {
        for (; keyList.isCons(); keyList = advanceCons(keyList))
            if (isEqual(key, getFirst(keyList)))
                return true;

        return false;
    }

    private SExpression advanceCons(SExpression knownCons) {
        return ((Cons) knownCons).getRest();
    }

    private SExpression getFirst(SExpression knownCons) {
        return ((Cons) knownCons).getFirst();
    }

    private SExpression evaluateConsequents(SExpression consequentList) {
        SExpression lastConsequentValue = Nil.INSTANCE;

        for (; consequentList.isCons(); consequentList = advanceCons(consequentList))
            lastConsequentValue = eval(getFirst(consequentList));

        return lastConsequentValue;
    }
}
